@@ -31,11 +31,22 @@ class ContentType(enum.Enum):
 class _RequestHandler:
     def __init__(self, credentials: CredentialsBase, configuration: Configuration) -> None:
         super().__init__()
+
         self.__credentials = credentials
         self.__oauth_service = OAuthService(configuration)
         self.__cookie = self.__oauth_service.login(self.__credentials)
-        self.__default_headers = self._get_default_headers()
         self.__configuration = configuration
+        self.__csrf_token = self._get_csrf_token()
+        self.__default_headers = self._get_default_headers()
+
+    def _get_csrf_token(self) -> str:
+        # Make a request to get the csrf token
+        hs_server_url = f'https://{self.__configuration.hs_domain}'
+        message = self._create_request_message(hs_server_url, [], 'GET', None)
+        self.__default_headers = self._get_default_headers(include_csrf=False)
+        res = self._send_request(message)
+        res.raise_for_status()
+        return res.cookies.get('csrftoken')
 
     def handle_request(
         self,
@@ -63,7 +74,9 @@ class _RequestHandler:
         res = self._send_request(message)
         if self._is_session_expired(res):
             self.__cookie = self.__oauth_service.login(self.__credentials)
+            self.__csrf_token = self._get_csrf_token()
             self.__default_headers = self._get_default_headers()
+            message = self._create_request_message(prepared_url, data, method, content_type)
             res = self._send_request(message)
 
         return res
@@ -94,13 +107,22 @@ class _RequestHandler:
 
     def _is_session_expired(self, res: Response) -> bool:
         redirect_header = res.headers.get('Location', '')
-        return res.status_code == 302 and self.__configuration.auth_server in redirect_header
+        is_cookie_expired = (
+            res.status_code == 302 and self.__configuration.auth_server in redirect_header
+        )
+        is_csrf_token_expired = res.status_code == 403 and 'CSRF Failed' in res.text
+        return is_cookie_expired or is_csrf_token_expired
 
-    def _get_default_headers(self) -> Dict[str, str]:
-        return {
+    def _get_default_headers(self, include_csrf: bool = True) -> Dict[str, str]:
+        headers = {
             'User-Agent': f'hyperscience-saas-client-python/{SAAS_CLIENT_VERSION}',
             COOKIE_HEADER: self.__cookie,
         }
+        if include_csrf and self.__csrf_token:
+            headers['X-CSRFToken'] = self.__csrf_token
+            headers['Referer'] = f'https://{self.__configuration.hs_domain}'
+            headers[COOKIE_HEADER] = f'csrftoken={self.__csrf_token}; {self.__cookie}'
+        return headers
 
     def _create_request_message(
         self,
